/******************************************************************************************************/
/*  Program: PROC S3 GET Files Macro Program.sas                                                      */
/*                                                                                                    */
/*  Author: Michael A. Raithel, Westat                                                                */
/*                                                                                                    */
/*  Created: 12/16/2021                                                                               */
/*                                                                                                    */
/*  Purpose: This program reads a flat file with the names of objects in an S3 bucket that are to be  */
/*           downloaded to a network drive. The program builds a PROC S3 "GET" statement for each     */
/*           object and executes PROC S3 to download the files to the specified network directory.    */
/*                                                                                                    */
/*  Parameters: Users must specify the following six parameters:                                      */
/*                                                                                                    */
/*             KEYID    - The access key ID for the S3 service                                        */
/*             REGION   - The region name for the S3 service                                          */
/*             SECRET   - The secret access key for the S3 service                                    */
/*             BUCKET   - The name of the AWS bucket folder the objects reside in                     */
/*             OUTDIR   - The full path name of the directory where the objects are to be downloaded  */
/*             FILELIST - The full path name of the flat file containing the names of the objects to  */
/*                        be downloaded                                                               */ 
/*                        NOTE: The program is expecting one object name per line within the flat file*/
/*                                                                                                    */
/*  Outputs: This program downloads the objects in the flat file specified by the FILELIST parameter  */
/*           to the directory specified by the OUTDIR parameter.                                      */
/*                                                                                                    */
/*           It also produces a PDF report file in the OUTDIR directory with up to three reports:     */
/*               1. Objects Successfully Dowloaded from S3                                            */
/*               2. Objects Listed In the Input File That Were NOT Downloaded from S3                 */
/*               3. Objects Existing in the Target Directory that Were Not in the Input List File     */
/*                                                                                                    */
/*           The name of the report file is:                                                          */
/*                                                                                                    */
/*               S3 File Transfer Report For Files Listed in -- <INPUTFILENAME>.pdf                   */
/*                                                                                                    */
/*  Change Log:                                                                                       */
/*                                                                                                    */
/******************************************************************************************************/

%Macro GETS3FIL(KEYID=, REGION=, SECRET=, BUCKET=, OUTDIR=, FILELIST=);

OPTIONS linesize=200 source2 msglevel=I ;
ODS Listing close;

/****************************************************************************************/
/* Part 1                                                                               */
/****************************************************************************************/
/* File allocations                                                                     */
/****************************************************************************************/

/* Allocate the file containing the list of objects to be retrieved from the S3 bucket */
filename filelist "&FILELIST";

/* Obtain the name of the input flat file to use as part of the output PDF report file name */
%LET INPUTFILENAME=%SCAN(&FILELIST,-2,.\);

/* Allocate the temporary file used to store the S3 statements this program builds */
filename getstmts temp;


/*****************************************************************************************/
/* Part 2                                                                                */
/*****************************************************************************************/
/* DATA step to build the PROC S3 statements, store them in the FILELIST temporary file, */
/* and then %INCLUDE the FILELIST temporary file in order to execute those S3 statements */
/* to actually download the specified objects from the S3 bucket.                        */
/*****************************************************************************************/

data buildgets(keep=ObjectName);

/* Read the records containing the names of files to be retrieved from S3 */
infile FILELIST length=linelen end=eof;
input ObjectName $varying200. linelen;

/* Specify the length of the statements to be stored in the temporary file */
length outcode $250.;

/* Specify the temporary file to be used by future PUT statements */
file getstmts;

QUOTE = '"';

/* This DO group builds the PROC S3 PROCEDURE statement and writes it to the temporary file */
if _n_ = 1 then do;

	outcode = "PROC S3 KEYID=" ||
			  QUOTE			   ||
 	          "&KEYID"         ||
			  QUOTE			   ||
			  " REGION="       ||
			  QUOTE			   ||
			  "&REGION"        ||
			  QUOTE			   ||
			  " SECRET="       ||
			  QUOTE			   ||
			  "&SECRET"        ||
			  QUOTE			   ||
			  ";"
	;
	put outcode;
end;

/* These statements build the PROC S3 GET statement and write it to the temporary file */
outcode = "GET " 			||
		  QUOTE				||
		  "/"               ||
		  "&BUCKET" 		||
		  "/"				||
		  strip(ObjectName) ||
		  QUOTE       		||
		  " "				||
		  QUOTE				||
		  "&OUTDIR"			||
		  "\"				||
		  strip(ObjectName) ||
		  QUOTE       		||
		  ";"
;
put outcode;

/* When the last record is processed, write the RUN statement to the temporary file */
if eof then do;

	outcode = "run;";
	put outcode;

end;

run;

/* %INCLUDE the temporary file to execute the perfectly formed S3 procedure */

%include getstmts;

/****************************************************************************************/
/* Part 3                                                                               */
/****************************************************************************************/
/* Prepare the data for the report(s) that are output in Part 4.                        */
/****************************************************************************************/

/* Perform DIR command on the output directory and pipe the results */
filename objects pipe "dir /b ""&OUTDIR\";

/* Read the names of the objects stored in the output directory */
data OutputDirFiles;

length ObjectName $200.;	   

infile objects length=linelen;
input ObjectName $varying200. linelen;

run;

/* Sort output directory object names for future merging */
proc sort data=OutputDirFiles;
	by ObjectName;
run;

/* Sort Buildgets for future merging */
proc sort data=BuildGets;
	by ObjectName;
run;


/* Merge the data sets together by ObjectName */
Data MatchingFiles
     BuildGetsOnly
	 OutputDirFilesOnly
	 ;
Merge BuildGets(in=a)
	  OutputDirFiles(in=B);

by ObjectName;

if a and b then output MatchingFiles;
if a and not b then output BuildGetsOnly;
if b and not a then output OutputDirFilesOnly;

run;

/****************************************************************************************/
/* Part 4                                                                               */
/****************************************************************************************/
/* Create PDF file with three possible report(s):                                       */
/*    1. Objects Successfully Dowloaded from S3                                         */
/*    2. Objects Listed In the Input File That Were NOT Downloaded from S3              */
/*    3. Objects Existing in the Target Directory that Were Not in the Input List File  */
/****************************************************************************************/

ods pdf file="&OUTDIR\S3 File Transfer Report For Files Listed in -- &INPUTFILENAME..pdf" style=daisy;

proc print data=MatchingFiles;
var ObjectName;
title1 "Files Successfully Downloaded from S3";
footnote1 "The Input File Containing Files to Download Was: &FILELIST";
footnote2 "The S3 Bucket Specified Was: &BUCKET";
run;

proc print data=BuildGetsOnly;
var ObjectName;
title1 "Files Listed In the Input File That Were NOT Downloaded from S3";
footnote1 "The Input File Containing Files to Download Was: &FILELIST";
footnote2 "The S3 Bucket Specified Was: &BUCKET";
run;

proc print data=OutputDirFilesOnly;
var ObjectName;
title1 "Files Existing in the Target Directory that Were Not in the Input List File";
footnote1 "The Input File Containing Files to Download Was: &FILELIST";
footnote2 "The S3 Bucket Specified Was: &BUCKET";
footnote3 "The Target Directory Was: &OUTDIR";
run;

ods pdf close;

%MEND GETS3FIL;




